/*	$NetBSD: exec_sun.c,v 1.7 1997/09/18 01:47:10 gwr Exp $ */

/*-
 * Copyright (c) 1982, 1986, 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * 	@(#)boot.c	8.1 (Berkeley) 6/10/93
 */

/* modified 7/28/99 for use with linux/sun3 -- Sam Creasey (sammy@oh.verio.com)
 */

#include <sys/param.h>
#include <machine/mon.h>
#include "bootinfo.h"
/* your linux build machine must provide this file */
#include "/usr/include/elf.h"

#include "stand.h"

extern void ICIA();
extern int debug;
int errno;

/* where in memory to stick the bootinfo records
(use 0x1000 before load, which is unused memory anyway */
#define BI_ADDR (loadaddr-0x1000)

#define CL_SIZE 256

/*ARGSUSED*/
int
exec_sun(cmdline, loadaddr)
	char	*cmdline;
	char	*loadaddr;
{
	register int io;
	Elf32_Ehdr x;
	Elf32_Phdr y;
	int cc, magic;
	void (*entry)();
	register char *cp;
	register int *ip;
	struct bi_record *bootinfo;
	unsigned long *bi_size;
	char *file;

	// 4 bytes for size, then the bootinfo
	bi_size = (unsigned long *)BI_ADDR;
	*bi_size = 0;
	bootinfo = (struct bi_record *)&bi_size[1];

#ifdef	DEBUG
	printf("exec_sun: cmdline=%s loadaddr=0x%x\n", cmdline, loadaddr);
#endif

	for(file = cmdline ; (*cmdline != ' ') && (*cmdline != 0); cmdline++);

	if(*cmdline != 0) {
	     *cmdline = 0;
	     cmdline++;
	}

	io = open(file, 0);
	if (io < 0)
		return(-1);

	/*
	 * Read in the exec header, and validate it.
	 */

	if (read(io, (char *)&x, sizeof(x)) != sizeof(x))
		goto shread;
	
	/* should be checking elf magic here! */
	
	entry = (void (*))loadaddr;
#ifdef DEBUG
	printf("e_phoff: %x\n", x.e_phoff);
	printf("e_ehsize: %x\n", x.e_ehsize);
#endif

	/* read in the program header */
	/* the first phdr containts all of what we card about... */
	lseek(io, SEEK_SET, x.e_phoff);
	if (read(io, (char *)&y, sizeof(y)) != sizeof(y))
		goto shread;

#ifdef DEBUG
	printf("p_offset: %x\n", y.p_offset);
	printf("p_filesz: %x\n", y.p_filesz);
	printf("p_memsz: %x\n", y.p_memsz);
#endif
	
	cp = (char *)entry;
	printf("text+data: %x", y.p_filesz);
	
	lseek(io, SEEK_SET, y.p_offset);
	if(read(io, cp, y.p_filesz) !=  y.p_filesz)
	     goto shread;

	cp += y.p_filesz;
	/* zero bss section */
	cc = y.p_memsz - y.p_filesz;
	printf(" bss: %x\n", cc);
	
	for(;cc;cc--)
	     *cp++ = 0;

	/* insert bootinfo structures for the kernel */
	if(*cmdline != 0) {
	     bootinfo->tag = BI_COMMAND_LINE;
	     bootinfo->size = sizeof(struct bi_record) + strlen(cmdline);
	     memcpy(bootinfo->data, cmdline, strlen(cmdline));
	     *bi_size += bootinfo->size;
	     bootinfo = (struct bi_record *)((unsigned long)bootinfo+
					     bootinfo->size);
	}
	*bi_size += sizeof(struct bi_record);
	bootinfo->tag = BI_LAST;
	bootinfo->size = sizeof(struct bi_record);
	
	*bi_size = (*bi_size/4)+1;
	close(io);

	printf("Starting program at 0x%x\n", (int)entry);
	asm("_exec_sun_call_entry:");
	ICIA();
	(*entry)();
	panic("exec returned");

shread:
	printf("exec: short read\n");
	errno = EIO;
closeout:
	close(io);
	return(-1);
}
